/*
	infoserver
	a demo of sockets programming with TCP/IP
	This server accepts requests for commands
	and returns results. Results may be multiple
	lines long.

	usage:	infoserver [-d] [-p port]
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/errno.h>   /* defines ERESTART, EINTR */
#include <signal.h>      /* for signal() */
#include <sys/wait.h>    /* defines WNOHANG, for wait() */

/* includes for user commands */
#include <unistd.h>
#include <sys/types.h>


#include "port.h"
#ifndef ERESTART	/* Not all systems recognize the ERESTART signal */
#define ERESTART EINTR
#endif

extern int errno;

int debug = 0;
void serve(int port);	/* main server function */
static void sigcld_handler(int);     /* signal handler */

main(int argc, char **argv)
{
	extern char *optarg;
	extern int optind;
	int c, err = 0; 
	int port = SERVICE_PORT;
	static char usage[] = "usage: %s [-d] [-p port] [string ...]\n";

	while ((c = getopt(argc, argv, "dp:")) != -1)
		switch (c) {
		case 'd':
			debug = 1;
			break;
		case 'p':
			port = atoi(optarg);
			if (port < 1024 || port > 65535) {
				fprintf(stderr, "invalid port number: %s\n", optarg);
				err = 1;
			}
			break;
		case '?':
			err = 1;
			break;
		}
	if (err || (optind < argc)) {
		fprintf(stderr, usage, argv[0]);
		exit(1);
	}
	serve(port);
}

/* serve: set up the service */

void
serve(int port)
{
	int svc;        /* listening socket providing service */
	int rqst;       /* socket accepting the request */
	socklen_t alen;       /* length of address structure */
	struct sockaddr_in my_addr;    /* address of this service */
	struct sockaddr_in client_addr;  /* client's address */
	int sockoptval = 1;
	char hostname[128]; /* host name, for debugging */

	gethostname(hostname, 128);

	/* catch SIGCHLD when a child exits */
	if (signal(SIGCHLD, (void (*)(int))sigcld_handler) == SIG_ERR) {
		perror("server-signal");
		exit(1);
	}

	/* get a tcp/ip socket */
	/*   AF_INET is the Internet address (protocol) family  */
	/*   with SOCK_STREAM we ask for a sequenced, reliable, two-way */
	/*   conenction based on byte streams.  With IP, this means that */
	/*   TCP will be used */

	if ((svc = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("cannot create socket");
		exit(1);
	}

	/* we use setsockopt to set SO_REUSEADDR. This allows us */
	/* to reuse the port immediately as soon as the service exits. */
	/* Some operating systems will not allow immediate reuse */
	/* on the chance that some packets may still be en route */
	/* to the port. */

	setsockopt(svc, SOL_SOCKET, SO_REUSEADDR, &sockoptval, sizeof(int));

	/* set up our address */
	/* htons converts a short integer into the network representation */
	/* htonl converts a long integer into the network representation */
	/* INADDR_ANY is the special IP address 0.0.0.0 which binds the */
	/* transport endpoint to all IP addresses on the machine. */

	memset((char*)&my_addr, 0, sizeof(my_addr));  /* 0 out the structure */
	my_addr.sin_family = AF_INET;   /* address family */
	my_addr.sin_port = htons(port);
	my_addr.sin_addr.s_addr = htonl(INADDR_ANY);

	/* bind to the address to which the service will be offered */
	if (bind(svc, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0) {
		perror("bind failed");
		exit(1);
	}

	/* set up the socket for listening with a queue length of 5 */
	if (listen(svc, 5) < 0) {
		perror("listen failed");
		exit(1);
	}

	if (debug)
		printf("server started on %s, listening on port %d\n",
			hostname, port);

	/* loop forever - wait for connection requests and perform the service */
	alen = sizeof(client_addr);     /* length of address */

	for (;;) {
		while ((rqst = accept(svc,
		                (struct sockaddr *)&client_addr, &alen)) < 0) {
			/* we may break out of accept if the system call */
			/* was interrupted. In this case, loop back and */
			/* try again */
			if ((errno != ECHILD) && (errno != ERESTART) && (errno != EINTR)) {
				perror("accept failed");
				exit(1);
			}
		}

		/* fork off a process to handle the client request */
		switch (fork()) {
		case -1:
			perror("fork failed\n");
			exit(1);

		default: /* parent */
			close(rqst);
			break;  /* continue looping for next connection */

		case 0:  /* child */
			close(svc);  /* child doesn't need this */
			dostuff(rqst); /* run the service */
			exit(0);
		}
	}
}

/* signal handler - note when a child process dies */

static void
sigcld_handler(int signum)
{
	int status;
	pid_t pid;

#ifdef LINUX
	signal(SIGCHLD, (void (*)(int))sigcld_handler);
#endif

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
		printf("server child %d exited with return value %d\n",
		        pid, WEXITSTATUS(status));
	}
}

/* dostuff: process client requests */

dostuff(int fd) {
	char *getdata(int);
	void do_cmd(int fd, char *cmd);

	char *cmd = "";
	char st[512];

	if (debug)
		printf("dostuff\n");
	while (cmd = getdata(fd)) {

		if (debug) printf("  got command \"%s\"\n", cmd);
		fflush(stdout);

		/* figure out what to do with the command */
		do_cmd(fd, cmd);
	}
}

/* convenience function - write a string to a file descriptor */
/* the write includes the terminating 0 */

void
print(int f, char *s)
{
	if (s == 0)
		write(f, "", 1);   /* send back a 0 byte */
	else
		write(f, s, strlen(s)+1);
}

/* commands that we execute on behalf of remote user */
/* all have the same interface: void cmd(int fd) */

/* functions for user commands */
void c_uptime(int);
void c_id(int);
void c_help(int);
void c_hello(int);
void c_bye(int);

/* table of commands that we support */

struct commands {
	char *name;
	void (*f)(int);
} cmdtab[] = {
	{ "hello", c_hello },
	{ "help", c_help },
	{ "bye", c_bye },
	{ "goodbye", c_bye },
	{ "uptime", c_uptime },
	{ "id", c_id },
	{ 0, 0 }
};

/* execute the appropriate command */
/* by looking it up in cmdtab and taking the */
/* appropriate action */

void
do_cmd(int fd, char *cmd) {
	int i;

	for (i=0; cmdtab[i].name; ++i) {
		if (!strcmp(cmdtab[i].name, cmd)) {
			cmdtab[i].f(fd);
			return;
		}
	}
	print(fd, "invalid command\n");
}

/* print a help msg */
void
c_help(int fd) {
	int i;
	FILE *f = fdopen(fd, "w");
	
	fprintf(f, "commands are: ");
	for (i=0; cmdtab[i].name; ++i) 
		fprintf(f, "%s ", cmdtab[i].name);
	fprintf(f, "\n");
	fflush(f); print(fd, ""); /* terminate the message */
	/* fclose(f); */
}

/* say hello */

void
c_hello(int fd)
{
	print(fd, "hi!\n");
}

/* say hello */

void
c_bye(int fd)
{
	print(fd, "nice knowing you...\n");
	exit(0);
}

/* return user ID */

void
c_id(int fd) {
	int uid = getuid();
	char buf[32];
	sprintf(buf, "user id=%d\n", uid);
	print(fd, buf);
}

/* return system uptime via uptime command */

void
c_uptime(int fd)
{
	int pid;
	switch (pid=fork()) {
	case -1:
		print(fd, "error forking command");
		break;

	default: /* parent */
		waitpid(pid, (int *)0, 0);
		if (debug) printf("c_uptime: child exited\n");
		write(fd, "", 1); /* write 0 to terminate output */
		break;

	case 0: /* child */
		close(0); /* close stdin, stdout, stderr */
		close(1);
		close(2);
		dup2(fd, 1);  /* duplicate socket to stdout */
		dup2(fd, 2);  /* duplicate socket to stderr */
		execlp("uptime", "uptime", 0);   /* run the command */
		/* if it failed... */
		{
			char *errmsg = "error running command";
			write(fd, errmsg, strlen(errmsg));
		}
		exit(0);
	}
}
