/*
	echoc: a demo of TCP/IP sockets programming
 	      echo client

	usage:	echoc [-d] [-h serverhost] [-p port]

	reads a line from stdin and sends it over the socket
	then waits for a return line from the socket.

*/

#include <stdio.h>
#include <stdlib.h>	/* needed for os x*/
#include <string.h>	/* for strlen */
#include <netdb.h>      /* for gethostbyname() */
#include <unistd.h>     /* for isatty() */
#include <sys/socket.h>
#include <netinet/in.h>

#include "echo.h"       /* defines default port */

int debug = 0;
int conn(char *host, int port);
void disconn();
void cmd(char *s);      /* process remote commands */

main(int argc, char **argv)
{
	extern char *optarg;
	extern int optind;
	int c, err = 0; 
	char *prompt = 0;
	int port = SERVICE_PORT;	/* default: whatever is in port.h */
	char *host = "localhost";	/* default: this host */
	char buf[256], *fgets();
	int blen;
	static char usage[] = 
	              "usage: %s [-d] [-h serverhost] [-p port]\n";

	if (isatty(0))         /* see if stdin is a terminal */
		prompt = ">";       /* if so, we'll display a prompt */

	while ((c = getopt(argc, argv, "dh:p:")) != -1)
		switch (c) {
		case 'd':  /* debug */
			debug = 1;
			break;
		case 'h':  /* hostname */
			host = optarg;
			break;
		case 'p':  /* port number */
			port = atoi(optarg);
			if (port < 1024 || port > 65535) {
				fprintf(stderr, "invalid port number: %s\n", optarg);
				err = 1;
			}
			break;
		case '?':
			err = 1;
			break;
		}
	if (err || (optind < argc)) {
		fprintf(stderr, usage, argv[0]);
		exit(1);
	}

	if (debug)
		printf("connecting to %s, port %d\n", host, port);

	if (!conn(host, port))    /* connect */
		exit(1);   /* something went wrong */

	/* read them from stdin - one per line */
	if (prompt) printf(prompt);
	while (fgets(buf, 256, stdin)) {
		blen=strlen(buf);
		/* strip off terminating newline */
		if (blen && buf[blen-1] == '\n')
			buf[blen-1] = '\0';
		if (echo(buf))
			break;
		if (prompt) printf(prompt);
	}
	disconn();    /* disconnect */
	return 0;
}

int fd;  /* fd is the file descriptor for the connected socket */

/* conn: connect to the service running on host:port */
/* return 0 on failure, non-zero on success */
int
conn(char *host, int port)
{
	struct hostent *hp;	/* host information */
	struct sockaddr_in myaddr;	/* our address */
	struct sockaddr_in servaddr;	/* server address */

	if (debug)
		printf("conn(host=\"%s\", port=\"%d\")\n",
		       host, port);

	/* get a tcp/ip socket */
	/* We do this as we did it for the server */
	/* request the Internet address protocol */
	/* and a reliable 2-way byte stream */

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("cannot create socket");
		return 0;
	}

	/* bind to an arbitrary return address */
	/* because this is the client side, we don't care about the address */
	/* no application will connect here  --- */
	/* INADDR_ANY is the IP address and 0 is the socket */
	/* htonl converts a long integer (e.g. address) to a network */
	/* representation (agreed-upon byte ordering */

	memset((char *)&myaddr, 0, sizeof(myaddr));
	myaddr.sin_family = AF_INET;
	myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	myaddr.sin_port = htons(0);

	if (bind(fd, (struct sockaddr *)&myaddr, sizeof(myaddr)) < 0) {
		perror("bind failed");
		return 0;
	}

	/* fill in the server's address and data */
	/* htons() converts a short integer to a network representation */

	memset((char*)&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(port);

	/* look up the address of the server given its name */
	hp = gethostbyname(host);
	if (!hp) {
		fprintf(stderr, "could not obtain address of %s\n", host);
		return 0;
	}

	/* put the host's address into the server address structure */
	memcpy((void *)&servaddr.sin_addr, hp->h_addr_list[0], hp->h_length);

	/* connect to server */
	if (connect(fd, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0) {
		perror("connect failed");
		return 0;
	}
	return 1;
}

/* disconnect from the service */
void
disconn(char *host, int port)
{
	if (debug)
		printf("disconn()\n");
	shutdown(fd, 2);    /* 2 means future sends & receives are disallowed */
}


/* send a string over the socket and print the returned string */
/* return 1 if socket got closed, 0 otherwise */
int
echo(char *s)
{
	int len;
	char *getdata(int fd);
	char buf[256];

	/* send it to the server */
	/* the command is terminated with a newline */
	len = strlen(s)+1;
	s[len-1] = '\n';

	if (debug)
		printf("writing %d bytes to socket\n", len);
	if (write(fd, s, len) != len) {
		fprintf(stderr, "write to socket failed\n");
		return;
	}
	/* get the return */
	if (getline(buf, 256, fd))
		puts(buf);
	else {
		/* the socket got closed - bail out */
		return 1;
	}
	return 0;
}
