/*
	server  a tiny demo of sockets programming
	  this one is connection-oriented
	  single threaded, single process version:
	  	only one client request at a time.

	usage:	server
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/errno.h>   /* defines ERESTART, EINTR */
#include <sys/wait.h>    /* defines WNOHANG, for wait() */
#ifdef __CYGWIN__	/* cygwin doesn't handle ERESTART */
#define ERESTART 99999	/* bogus value */
#endif
#define ERESTART 99999	/* bogus value */

extern int errno;

void serve(int port);	/* main server function */
char *ipfmt(unsigned long a);	/* format address to printable format */

main(int argc, char **argv)
{
	char *usage = "usage: server port\n";
	int port;

	if (argc != 2) {
		fprintf(stderr, usage);
		exit(1);
	}
	port = atoi(argv[1]);
	if (port <= 0 || port > 65535) {
		fprintf(stderr, "invalid port: %s\n", argv[1]);
		fprintf(stderr, usage);
		exit(1);
	}
	serve(port);
}

/* serve: set up the service */

void
serve(int port)
{
	int svc;        /* listening socket providing service */
	int rqst;       /* socket accepting the request */
	socklen_t alen;       /* length of address structure */
	struct sockaddr_in my_addr;    /* address of this service */
	struct sockaddr_in client_addr;  /* client's address */
	int sockoptval = 1;
	char hostname[128]; /* host name, for debugging */

	gethostname(hostname, 128);

	/* get a tcp/ip socket */
	/*   AF_INET is the Internet address (protocol) family  */
	/*   with SOCK_STREAM we ask for a sequenced, reliable, two-way */
	/*   conenction based on byte streams.  With IP, this means that */
	/*   TCP will be used */

	if ((svc = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("cannot create socket");
		exit(1);
	}

	/* we use setsockopt to set SO_REUSEADDR. This allows us */
	/* to reuse the port immediately as soon as the service exits. */
	/* Some operating systems will not allow immediate reuse */
	/* on the chance that some packets may still be en route */
	/* to the port. */

	setsockopt(svc, SOL_SOCKET, SO_REUSEADDR, &sockoptval, sizeof(int));

	/* set up our address */
	/* htons converts a short integer into the network representation */
	/* htonl converts a long integer into the network representation */
	/* INADDR_ANY is the special IP address 0.0.0.0 which binds the */
	/* transport endpoint to all IP addresses on the machine. */

	memset((void*)&my_addr, 0, sizeof(my_addr));  /* 0 out the structure */
	my_addr.sin_family = AF_INET;   /* address family */
	my_addr.sin_port = htons(port);
	my_addr.sin_addr.s_addr = htonl(INADDR_ANY);

	/* bind to the address to which the service will be offered */

	if (bind(svc, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0) {
		perror("bind failed");
		exit(1);
	}

	/* set up the socket for listening with a queue length of 5 */
	if (listen(svc, 5) < 0) {
		perror("listen failed");
		exit(1);
	}

	printf("server started on %s, listening on port %d\n", hostname, port);

	/* loop forever - wait for connection requests and perform the service */
	alen = sizeof(client_addr);     /* length of address */

	for (;;) {
		while ((rqst = accept(svc,
	                (struct sockaddr *)&client_addr, &alen)) < 0) {
			/* we may break out of accept if the system call */
			/* was interrupted. In this case, loop back and */
			/* try again */
			if ((errno != ERESTART) && (errno != EINTR)) {
				perror("accept failed");
				exit(1);
			}
		}
		/* we got a request */

		/* print some debugging output 
		 * note the use of ntohl to convert a 32-bit quantity (long)
		 * from the network representation to the local (host)
		 * representation. ntohs (network to host short) does the
		 * same for 16-bit quantities
		 */
		printf("received request from %s port %d\n",
			ipfmt(ntohl(client_addr.sin_addr.s_addr)),
			ntohs(client_addr.sin_port));

		dostuff(rqst); /* run the service */
		printf("connection closed\n");
		close(rqst);
	}
}

/* dostuff: process client requests */
/* here we just read from the client until the client closes the connection */
/* all data is printed to the standard output */

dostuff(int fd) {
	char buf[BUFSIZ];
	int n;

	while ((n=read(fd, buf, BUFSIZ)) > 0)
		write(1, buf, n);
}

/* format ip address to dot-separated text format */

char *
ipfmt(unsigned long a) {
	static char p[16];
	sprintf(p, "%u.%u.%u.%u", (a>>24)&0xff, (a>>16)&0xff,
	        (a>>8)&0xff, a&0xff);
	return p;
}
